/* Preprocessor for text to pretty-print it in Postscript. */
#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#define MAXDEPTH 8

main(ac, av)
int ac;
char **av;
{
	FILE *fp;
	char buffer[512];
	char *t, *ctime();
	struct stat sb;

	if(ac>1) {
		while(*++av) {
			if(fp = fopen(*av, "r")) {
				if(fstat(fileno(fp), &sb) == 0) {
					t = ctime(&sb.st_mtime);
					t[strlen(t)-1] = 0;
					printf("/RTEXT (%s) def\n", t);
					printf("/RFONT ROMAN def\n");
				} else {
					printf("/RTEXT DATE def\n");
					printf("/RFONT ITALIC def\n");
				}
				printf("/LTEXT (%s) def\n", *av);
				printf("/LFONT ROMAN def\n");
				printf("BOF\n");
				initfile();
				while(fgets(buffer, sizeof(buffer), fp)) {
					puttext(buffer);
					putchar('\n');
				}
				printf("EOF\n");
			}
		}
	} else {
		printf("/RTEXT DATE def\n");
		printf("/RFONT ITALIC def\n");
		printf("/LTEXT TITLE def\n", *av);
		printf("/LFONT ITALIC def\n");
		printf("BOF\n");
		initfile();
		while(fgets(buffer, sizeof(buffer), stdin)) {
			puttext(buffer);
			putchar('\n');
		}
		printf("EOF\n");
	}
}

char workbuf[MAXDEPTH][80];
int depth[80];
int style[80];
#define NORMAL 0
#define BOLD 1
#define ITALIC 2
#define UNKNOWN 3
char   *stylename[4] = {
	"NORMAL ", "BOLD ", "SLANT ", "(Can't Happen!) print "
};
int oldstyle;
int col;

initfile()
{
	oldstyle = UNKNOWN;
}

initbuf()
{
	int i;

	for(col = 0; col < 80; col++) {
		depth[col] = 0;
		style[col] = UNKNOWN;
		for(i = 0; i < MAXDEPTH; i++) {
			workbuf[i][col] = 0;
		}
	}

	col = 0;
}

begintext()
{
	printf("(");
}

endtext()
{
	printf(") P ");
}

printab(col)
int col;
{
	printf("%d T ", col);
}

printos(col)
int col;
{
	int i;
	printf("[ ");
	for(i = 0; i < depth[col]; i++) {
		printf("(");
		printchar(col, i);
		printf(") ");
	}
	printf("] OS ");
}

printchar(col, d)
int col, d;
{
	char c = workbuf[d][col];

	if(c < 0x80) {
		if(strchr("()\\", c))
			putchar('\\');
		putchar(c);
	}
	else printf("\\%03o", c);
}

patch_style()
{
	int col, i;

	for(col = 0; col < 80; col++) {
		if(workbuf[0][col] == ' ' || depth[col] == 0) {
			if(col == 0) {
				if(oldstyle == UNKNOWN)
					style[col] = NORMAL;
				else
					style[col] = oldstyle;
			} else {
				if(style[col-1] == UNKNOWN)
					style[col] = NORMAL;
				else
					style[col] = style[col-1];
			}
		} else if(workbuf[0][col] == '_' && style[col] == ITALIC) {
			if(col == 0 || style[col-1] != ITALIC)
				style[col] = BOLD;
		} else if(depth[col] > 1) {
			for(i = 1; i < depth[col]; i++)
				if(workbuf[i][col] != workbuf[i-1][col])
					break;
			if(i == depth[col]) {
				depth[col] = 1;
				style[col] = BOLD;
			}
		}
	}
}

endline()
{
	int intab = 0;
	int intext = 0;
	int col;
	int newstyle;

	patch_style();
	printf("NL ");
	for(col = 0; col < 80; col++) {
		if(style[col] == UNKNOWN)
			newstyle = NORMAL;
		else
			newstyle = style[col];
		switch(depth[col]) {
			case 0:
				if(intext) {
					endtext();
					intext = 0;
				}
				intab = 1;
				break;
			case 1: 
				if(intab) {
					printab(col);
					intab = 0;
				}
				if(oldstyle != newstyle) {
					if(intext) endtext();
					printf(stylename[newstyle]);
					begintext();
					intext = 1;
					oldstyle = newstyle;
				}
				else if(!intext) {
					begintext();
					intext = 1;
				}
				printchar(col, 0);
				break;
			default:
				if(intab) {
					printab(col);
					intab = 0;
				}
				if(intext) {
					endtext();
					intext = 0;
				}
				if(oldstyle != newstyle) {
					printf(newstyle);
					oldstyle = newstyle;
				}
				printos(col);
				break;
		}
	}
	if(intext)
		endtext();
}

pchar(c)
char c;
{
	int i;

	if(col>=80) {
		endline();
		initbuf();
	}

	switch(c) {
	    case '_':
		if(depth[col] == 0) {
			depth[col] = 1;
			workbuf[0][col] = '_';
		} else if(depth[col] == 1) {
			if(style[col] == NORMAL || style[col] == UNKNOWN)
				style[col] = ITALIC;
		}
		break;
	    case ' ':
		if(depth[col] == 0) {
			depth[col] = 1;
			workbuf[0][col] = ' ';
		}
		break;
	    default:
		switch(workbuf[0][col]) {
		    case '_':
			style[col] = ITALIC;
		    case ' ':
			workbuf[0][col] = c;
			break;
		    default:
			if(depth[col] < MAXDEPTH)
				workbuf[depth[col]++][col] = c;
			if(style[col] == UNKNOWN)
				style[col] = NORMAL;
			break;
		}
		break;
	}

	col++;
}

ptab()
{
	col += 8 - col%8;
}

pcr()
{
	col = 0;
}

pbs()
{
	if(col>0) col--;
}

pff()
{
	endline();
	printf("FF\n");
}

puttext(s)
char *s;
{
	initbuf();

	while(*s) {
		switch(*s) {
			case '\b': pbs(); break;
			case '\t': ptab(); break;
			case '\r': pcr(); break;
			case '\f': pff(); break;
			default:
				if(isprint(*s&0x7F))
					pchar(*s);
				break;
		}
		s++;
	}
	endline();
}
